package tomo

import "fmt"
import "git.tebibyte.media/tomo/tomo/event"

// Role describes the role of an object.
type Role struct {
	// Package is an optional namespace field. If specified, it should be
	// the package name or module name the object is from.
	Package string

	// Object specifies what type of object it is. For example:
	//   - TextInput
	//   - Table
	//   - Label
	//   - Dial
	// This should correspond directly to the type name of the object.
	Object string

	// Variant is an optional field to be used when an object has one or
	// more soft variants under one type. For example, an object "Slider"
	// may have variations "horizontal" and "vertical".
	Variant string
}

// String satisfies the fmt.Stringer interface.
// It follows the format of:
// Package.Object[Variant]
func (r Role) String () string {
	return fmt.Sprintf("%s.%s[%s]", r.Package, r.Object, r.Variant)
}

// R is shorthand for creating a Role structure.
func R (pack, object, variant string) Role {
	return Role { Package: pack, Object: object, Variant: variant }
}

// Color represents a color ID.
type Color int; const (
	ColorBackground Color = iota
	ColorForeground
	ColorRaised
	ColorSunken
	ColorAccent
)

// String satisfies the fmt.Stringer interface.
func (c Color) String () string {
	switch c {
	case ColorBackground: return "background"
	case ColorForeground: return "foreground"
	case ColorRaised:     return "raised"
	case ColorSunken:     return "sunken"
	case ColorAccent:     return "accent"
	default:              return "unknown"
	}
}

// RGBA satisfies the color.Color interface.
func (id Color) RGBA () (r, g, b, a uint32) {
	if style == nil { return }
	return style.RGBA(id)
}

// Style can apply a visual style to different objects.
type Style interface {
	// Apply applies the theme to the given object, according to its role.
	// This may register event listeners with the given object; closing the
	// returned cookie must remove them.
	Apply (Object) event.Cookie

	// RGBA returns the RGBA values of the corresponding color ID.
	RGBA (Color) (r, g, b, a uint32)
}

var style Style

// SetStyle sets the style.
func SetStyle (sty Style) {
	assertBackend()
	style = sty
	backend.SetStyle(sty)
}
